@@ -1,7 +1,15 @@
+/*
+ * $PostgreSQL: pgsql/contrib/btree_gist/btree_numeric.c,v 1.13 2009/06/11 14:48:50 momjian Exp $
+ */
 #include "btree_gist.h"
+
+#include <math.h>
+#include <float.h>
+
 #include "btree_utils_var.h"
 #include "utils/builtins.h"
 #include "utils/numeric.h"
+#include "utils/rel.h"
 
 /*
 ** Bytea ops
@@ -95,14 +103,19 @@ gbt_numeric_compress(PG_FUNCTION_ARGS)
 Datum
 gbt_numeric_consistent(PG_FUNCTION_ARGS)
 {
-
 	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
-	GBT_VARKEY *key = (GBT_VARKEY *) DatumGetPointer(entry->key);
 	void	   *query = (void *) DatumGetNumeric(PG_GETARG_DATUM(1));
 	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
+
+	/* Oid		subtype = PG_GETARG_OID(3); */
+	bool	   *recheck = (bool *) PG_GETARG_POINTER(4);
 	bool		retval;
+	GBT_VARKEY *key = (GBT_VARKEY *) DatumGetPointer(entry->key);
 	GBT_VARKEY_R r = gbt_var_key_readable(key);
 
+	/* All cases served by this function are exact */
+	*recheck = false;
+
 	retval = gbt_var_consistent(&r, query, &strategy, GIST_LEAF(entry), &tinfo);
 	PG_RETURN_BOOL(retval);
 }
