@@ -10,7 +10,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/libpq/hba.c,v 1.188 2009/06/24 13:39:42 mha Exp $
+ *	  $PostgreSQL: pgsql/src/backend/libpq/hba.c,v 1.188.2.5 2010/03/24 17:05:51 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -91,6 +91,10 @@ pg_isblank(const char c)
  * double quotes (this allows the inclusion of blanks, but not newlines).
  *
  * The token, if any, is returned at *buf (a buffer of size bufsz).
+ * Also, we set *initial_quote to indicate whether there was quoting before
+ * the first character.  (We use that to prevent "@x" from being treated
+ * as a file inclusion request.  Note that @"x" should be so treated;
+ * we want to allow that to support embedded spaces in file paths.)
  *
  * If successful: store null-terminated token at *buf and return TRUE.
  * If no more tokens on line: set *buf = '\0' and return FALSE.
@@ -105,7 +109,7 @@ pg_isblank(const char c)
  * token.
  */
 static bool
-next_token(FILE *fp, char *buf, int bufsz)
+next_token(FILE *fp, char *buf, int bufsz, bool *initial_quote)
 {
 	int			c;
 	char	   *start_buf = buf;
@@ -114,8 +118,11 @@ next_token(FILE *fp, char *buf, int bufsz)
 	bool		was_quote = false;
 	bool		saw_quote = false;
 
+	/* end_buf reserves two bytes to ensure we can append \n and \0 */
 	Assert(end_buf > start_buf);
 
+	*initial_quote = false;
+
 	/* Move over initial whitespace and commas */
 	while ((c = getc(fp)) != EOF && (pg_isblank(c) || c == ','))
 		;
@@ -174,6 +181,8 @@ next_token(FILE *fp, char *buf, int bufsz)
 		{
 			in_quote = !in_quote;
 			saw_quote = true;
+			if (buf == start_buf)
+				*initial_quote = true;
 		}
 
 		c = getc(fp);
@@ -216,12 +225,13 @@ next_token_expand(const char *filename, FILE *file)
 	char	   *comma_str = pstrdup("");
 	bool		got_something = false;
 	bool		trailing_comma;
+	bool		initial_quote;
 	char	   *incbuf;
 	int			needed;
 
 	do
 	{
-		if (!next_token(file, buf, sizeof(buf)))
+		if (!next_token(file, buf, sizeof(buf), &initial_quote))
 			break;
 
 		got_something = true;
@@ -235,7 +245,7 @@ next_token_expand(const char *filename, FILE *file)
 			trailing_comma = false;
 
 		/* Is this referencing a file? */
-		if (buf[0] == '@')
+		if (!initial_quote && buf[0] == '@' && buf[1] != '\0')
 			incbuf = tokenize_inc_file(filename, buf + 1);
 		else
 			incbuf = pstrdup(buf);
@@ -404,7 +414,7 @@ tokenize_file(const char *filename, FILE *file,
 
 	*lines = *line_nums = NIL;
 
-	while (!feof(file))
+	while (!feof(file) && !ferror(file))
 	{
 		buf = next_token_expand(filename, file);
 
@@ -710,6 +720,8 @@ parse_hba_line(List *line, int line_num, HbaLine *parsedline)
 								line_num, HbaFileName)));
 			return false;
 		}
+
+		/* need a modifiable copy of token */
 		token = pstrdup(lfirst(line_item));
 
 		/* Check if it has a CIDR suffix and if so isolate it */
@@ -736,16 +748,12 @@ parse_hba_line(List *line, int line_num, HbaLine *parsedline)
 							token, gai_strerror(ret)),
 					 errcontext("line %d of configuration file \"%s\"",
 								line_num, HbaFileName)));
-			if (cidr_slash)
-				*cidr_slash = '/';
 			if (gai_result)
 				pg_freeaddrinfo_all(hints.ai_family, gai_result);
+			pfree(token);
 			return false;
 		}
 
-		if (cidr_slash)
-			*cidr_slash = '/';
-
 		memcpy(&parsedline->addr, gai_result->ai_addr, gai_result->ai_addrlen);
 		pg_freeaddrinfo_all(hints.ai_family, gai_result);
 
@@ -755,18 +763,22 @@ parse_hba_line(List *line, int line_num, HbaLine *parsedline)
 			if (pg_sockaddr_cidr_mask(&parsedline->mask, cidr_slash + 1,
 									  parsedline->addr.ss_family) < 0)
 			{
+				*cidr_slash = '/';			/* restore token for message */
 				ereport(LOG,
 						(errcode(ERRCODE_CONFIG_FILE_ERROR),
 						 errmsg("invalid CIDR mask in address \"%s\"",
 								token),
 						 errcontext("line %d of configuration file \"%s\"",
 									line_num, HbaFileName)));
+				pfree(token);
 				return false;
 			}
+			pfree(token);
 		}
 		else
 		{
 			/* Read the mask field. */
+			pfree(token);
 			line_item = lnext(line_item);
 			if (!line_item)
 			{
@@ -910,6 +922,22 @@ parse_hba_line(List *line, int line_num, HbaLine *parsedline)
 		return false;
 	}
 
+	if (parsedline->conntype == ctLocal &&
+		parsedline->auth_method == uaGSS)
+	{
+		ereport(LOG,
+				(errcode(ERRCODE_CONFIG_FILE_ERROR),
+			 errmsg("gssapi authentication is not supported on local sockets"),
+				 errcontext("line %d of configuration file \"%s\"",
+							line_num, HbaFileName)));
+		return false;
+	}
+	/*
+	 * SSPI authentication can never be enabled on ctLocal connections, because
+	 * it's only supported on Windows, where ctLocal isn't supported.
+	 */
+
+
 	if (parsedline->conntype != ctHostSSL &&
 		parsedline->auth_method == uaCert)
 	{
@@ -1253,7 +1281,7 @@ load_role(void)
 }
 
 /*
- * Free the contents of a hba record
+ * Free an HbaLine structure
  */
 static void
 free_hba_record(HbaLine *record)
@@ -1262,6 +1290,8 @@ free_hba_record(HbaLine *record)
 		pfree(record->database);
 	if (record->role)
 		pfree(record->role);
+	if (record->usermap)
+		pfree(record->usermap);
 	if (record->pamservice)
 		pfree(record->pamservice);
 	if (record->ldapserver)
@@ -1274,6 +1304,7 @@ free_hba_record(HbaLine *record)
 		pfree(record->krb_server_hostname);
 	if (record->krb_realm)
 		pfree(record->krb_realm);
+	pfree(record);
 }
 
 /*
@@ -1342,20 +1373,22 @@ load_hba(void)
 		{
 			/* Parse error in the file, so indicate there's a problem */
 			free_hba_record(newline);
-			pfree(newline);
+			ok = false;
 
 			/*
 			 * Keep parsing the rest of the file so we can report errors on
 			 * more than the first row. Error has already been reported in the
 			 * parsing function, so no need to log it here.
 			 */
-			ok = false;
 			continue;
 		}
 
 		new_parsed_lines = lappend(new_parsed_lines, newline);
 	}
 
+	/* Free the temporary lists */
+	free_lines(&hba_lines, &hba_line_nums);
+
 	if (!ok)
 	{
 		/* Parsing failed at one or more rows, so bail out */
@@ -1367,9 +1400,6 @@ load_hba(void)
 	clean_hba_list(parsed_hba_lines);
 	parsed_hba_lines = new_parsed_lines;
 
-	/* Free the temporary lists */
-	free_lines(&hba_lines, &hba_line_nums);
-
 	return true;
 }
 
@@ -1392,28 +1422,29 @@ read_pg_database_line(FILE *fp, char *dbname, Oid *dboid,
 					  Oid *dbtablespace, TransactionId *dbfrozenxid)
 {
 	char		buf[MAX_TOKEN];
+	bool		initial_quote;
 
 	if (feof(fp))
 		return false;
-	if (!next_token(fp, buf, sizeof(buf)))
+	if (!next_token(fp, buf, sizeof(buf), &initial_quote))
 		return false;
 	if (strlen(buf) >= NAMEDATALEN)
 		elog(FATAL, "bad data in flat pg_database file");
 	strcpy(dbname, buf);
-	next_token(fp, buf, sizeof(buf));
+	next_token(fp, buf, sizeof(buf), &initial_quote);
 	if (!isdigit((unsigned char) buf[0]))
 		elog(FATAL, "bad data in flat pg_database file");
 	*dboid = atooid(buf);
-	next_token(fp, buf, sizeof(buf));
+	next_token(fp, buf, sizeof(buf), &initial_quote);
 	if (!isdigit((unsigned char) buf[0]))
 		elog(FATAL, "bad data in flat pg_database file");
 	*dbtablespace = atooid(buf);
-	next_token(fp, buf, sizeof(buf));
+	next_token(fp, buf, sizeof(buf), &initial_quote);
 	if (!isdigit((unsigned char) buf[0]))
 		elog(FATAL, "bad data in flat pg_database file");
 	*dbfrozenxid = atoxid(buf);
 	/* expect EOL next */
-	if (next_token(fp, buf, sizeof(buf)))
+	if (next_token(fp, buf, sizeof(buf), &initial_quote))
 		elog(FATAL, "bad data in flat pg_database file");
 	return true;
 }
@@ -1609,13 +1640,13 @@ ident_syntax:
 /*
  *	Scan the (pre-parsed) ident usermap file line by line, looking for a match
  *
- *	See if the user with ident username "ident_user" is allowed to act
- *	as Postgres user "pgrole" according to usermap "usermap_name".
+ *	See if the user with ident username "auth_user" is allowed to act
+ *	as Postgres user "pg_role" according to usermap "usermap_name".
  *
  *	Special case: Usermap NULL, equivalent to what was previously called
- *	"sameuser" or "samerole", don't look in the usermap
- *	file.  That's an implied map where "pgrole" must be identical to
- *	"ident_user" in order to be authorized.
+ *	"sameuser" or "samerole", means don't look in the usermap file.
+ *	That's an implied map wherein "pg_role" must be identical to
+ *	"auth_user" in order to be authorized.
  *
  *	Iff authorized, return STATUS_OK, otherwise return STATUS_ERROR.
  */
@@ -1642,7 +1673,7 @@ check_usermap(const char *usermap_name,
 		}
 		ereport(LOG,
 				(errmsg("provided username (%s) and authenticated username (%s) don't match",
-						auth_user, pg_role)));
+						pg_role, auth_user)));
 		return STATUS_ERROR;
 	}
 	else
