@@ -1,10 +1,10 @@
 /*
  * This file contains public functions for conversion between
- * client encoding and server internal encoding.
- * (currently mule internal code (mic) is used)
+ * client encoding and server (database) encoding.
+ *
  * Tatsuo Ishii
  *
- * $PostgreSQL: pgsql/src/backend/utils/mb/mbutils.c,v 1.69 2008/01/09 23:43:54 tgl Exp $
+ * $PostgreSQL: pgsql/src/backend/utils/mb/mbutils.c,v 1.87 2009/06/11 14:49:05 momjian Exp $
  */
 #include "postgres.h"
 
@@ -13,6 +13,7 @@
 #include "mb/pg_wchar.h"
 #include "utils/builtins.h"
 #include "utils/memutils.h"
+#include "utils/pg_locale.h"
 #include "utils/syscache.h"
 
 /*
@@ -27,23 +28,37 @@
 #define MAX_CONVERSION_GROWTH  4
 
 /*
- * We handle for actual FE and BE encoding setting encoding-identificator
- * and encoding-name too. It prevent searching and conversion from encoding
- * to encoding name in getdatabaseencoding() and other routines.
+ * We maintain a simple linked list caching the fmgr lookup info for the
+ * currently selected conversion functions, as well as any that have been
+ * selected previously in the current session.	(We remember previous
+ * settings because we must be able to restore a previous setting during
+ * transaction rollback, without doing any fresh catalog accesses.)
+ *
+ * Since we'll never release this data, we just keep it in TopMemoryContext.
+ */
+typedef struct ConvProcInfo
+{
+	int			s_encoding;		/* server and client encoding IDs */
+	int			c_encoding;
+	FmgrInfo	to_server_info; /* lookup info for conversion procs */
+	FmgrInfo	to_client_info;
+} ConvProcInfo;
+
+static List *ConvProcList = NIL;	/* List of ConvProcInfo */
+
+/*
+ * These variables point to the currently active conversion functions,
+ * or are NULL when no conversion is needed.
+ */
+static FmgrInfo *ToServerConvProc = NULL;
+static FmgrInfo *ToClientConvProc = NULL;
+
+/*
+ * These variables track the currently selected FE and BE encodings.
  */
 static pg_enc2name *ClientEncoding = &pg_enc2name_tbl[PG_SQL_ASCII];
 static pg_enc2name *DatabaseEncoding = &pg_enc2name_tbl[PG_SQL_ASCII];
 
-/*
- * Caches for conversion function info. These values are allocated in
- * MbProcContext. That context is a child of TopMemoryContext,
- * which allows these values to survive across transactions. See
- * SetClientEncoding() for more details.
- */
-static MemoryContext MbProcContext = NULL;
-static FmgrInfo *ToServerConvProc = NULL;
-static FmgrInfo *ToClientConvProc = NULL;
-
 /*
  * During backend startup we can't set client encoding because we (a)
  * can't look up the conversion functions, and (b) may not know the database
@@ -69,11 +84,7 @@ int
 SetClientEncoding(int encoding, bool doit)
 {
 	int			current_server_encoding;
-	Oid			to_server_proc,
-				to_client_proc;
-	FmgrInfo   *to_server;
-	FmgrInfo   *to_client;
-	MemoryContext oldcontext;
+	ListCell   *lc;
 
 	if (!PG_VALID_FE_ENCODING(encoding))
 		return -1;
@@ -100,79 +111,117 @@ SetClientEncoding(int encoding, bool doit)
 			ClientEncoding = &pg_enc2name_tbl[encoding];
 			ToServerConvProc = NULL;
 			ToClientConvProc = NULL;
-			if (MbProcContext)
-				MemoryContextReset(MbProcContext);
 		}
 		return 0;
 	}
 
-	/*
-	 * If we're not inside a transaction then we can't do catalog lookups, so
-	 * fail.  After backend startup, this could only happen if we are
-	 * re-reading postgresql.conf due to SIGHUP --- so basically this just
-	 * constrains the ability to change client_encoding on the fly from
-	 * postgresql.conf.  Which would probably be a stupid thing to do anyway.
-	 */
-	if (!IsTransactionState())
-		return -1;
-
-	/*
-	 * Look up the conversion functions.
-	 */
-	to_server_proc = FindDefaultConversionProc(encoding,
-											   current_server_encoding);
-	if (!OidIsValid(to_server_proc))
-		return -1;
-	to_client_proc = FindDefaultConversionProc(current_server_encoding,
-											   encoding);
-	if (!OidIsValid(to_client_proc))
-		return -1;
-
-	/*
-	 * Done if not wanting to actually apply setting.
-	 */
-	if (!doit)
-		return 0;
-
-	/* Before loading the new fmgr info, remove the old info, if any */
-	ToServerConvProc = NULL;
-	ToClientConvProc = NULL;
-	if (MbProcContext != NULL)
+	if (IsTransactionState())
 	{
-		MemoryContextReset(MbProcContext);
+		/*
+		 * If we're in a live transaction, it's safe to access the catalogs,
+		 * so look up the functions.  We repeat the lookup even if the info is
+		 * already cached, so that we can react to changes in the contents of
+		 * pg_conversion.
+		 */
+		Oid			to_server_proc,
+					to_client_proc;
+		ConvProcInfo *convinfo;
+		MemoryContext oldcontext;
+
+		to_server_proc = FindDefaultConversionProc(encoding,
+												   current_server_encoding);
+		if (!OidIsValid(to_server_proc))
+			return -1;
+		to_client_proc = FindDefaultConversionProc(current_server_encoding,
+												   encoding);
+		if (!OidIsValid(to_client_proc))
+			return -1;
+
+		/*
+		 * Done if not wanting to actually apply setting.
+		 */
+		if (!doit)
+			return 0;
+
+		/*
+		 * Load the fmgr info into TopMemoryContext (could still fail here)
+		 */
+		convinfo = (ConvProcInfo *) MemoryContextAlloc(TopMemoryContext,
+													   sizeof(ConvProcInfo));
+		convinfo->s_encoding = current_server_encoding;
+		convinfo->c_encoding = encoding;
+		fmgr_info_cxt(to_server_proc, &convinfo->to_server_info,
+					  TopMemoryContext);
+		fmgr_info_cxt(to_client_proc, &convinfo->to_client_info,
+					  TopMemoryContext);
+
+		/* Attach new info to head of list */
+		oldcontext = MemoryContextSwitchTo(TopMemoryContext);
+		ConvProcList = lcons(convinfo, ConvProcList);
+		MemoryContextSwitchTo(oldcontext);
+
+		/*
+		 * Everything is okay, so apply the setting.
+		 */
+		ClientEncoding = &pg_enc2name_tbl[encoding];
+		ToServerConvProc = &convinfo->to_server_info;
+		ToClientConvProc = &convinfo->to_client_info;
+
+		/*
+		 * Remove any older entry for the same encoding pair (this is just to
+		 * avoid memory leakage).
+		 */
+		foreach(lc, ConvProcList)
+		{
+			ConvProcInfo *oldinfo = (ConvProcInfo *) lfirst(lc);
+
+			if (oldinfo == convinfo)
+				continue;
+			if (oldinfo->s_encoding == convinfo->s_encoding &&
+				oldinfo->c_encoding == convinfo->c_encoding)
+			{
+				ConvProcList = list_delete_ptr(ConvProcList, oldinfo);
+				pfree(oldinfo);
+				break;			/* need not look further */
+			}
+		}
+
+		return 0;				/* success */
 	}
 	else
 	{
 		/*
-		 * This is the first time through, so create the context. Make it a
-		 * child of TopMemoryContext so that these values survive across
-		 * transactions.
+		 * If we're not in a live transaction, the only thing we can do is
+		 * restore a previous setting using the cache.	This covers all
+		 * transaction-rollback cases.	The only case it might not work for is
+		 * trying to change client_encoding on the fly by editing
+		 * postgresql.conf and SIGHUP'ing.  Which would probably be a stupid
+		 * thing to do anyway.
 		 */
-		MbProcContext = AllocSetContextCreate(TopMemoryContext,
-											  "MbProcContext",
-											  ALLOCSET_SMALL_MINSIZE,
-											  ALLOCSET_SMALL_INITSIZE,
-											  ALLOCSET_SMALL_MAXSIZE);
+		foreach(lc, ConvProcList)
+		{
+			ConvProcInfo *oldinfo = (ConvProcInfo *) lfirst(lc);
+
+			if (oldinfo->s_encoding == current_server_encoding &&
+				oldinfo->c_encoding == encoding)
+			{
+				if (doit)
+				{
+					ClientEncoding = &pg_enc2name_tbl[encoding];
+					ToServerConvProc = &oldinfo->to_server_info;
+					ToClientConvProc = &oldinfo->to_client_info;
+				}
+				return 0;
+			}
+		}
+
+		return -1;				/* it's not cached, so fail */
 	}
-
-	/* Load the fmgr info into MbProcContext */
-	oldcontext = MemoryContextSwitchTo(MbProcContext);
-	to_server = palloc(sizeof(FmgrInfo));
-	to_client = palloc(sizeof(FmgrInfo));
-	fmgr_info(to_server_proc, to_server);
-	fmgr_info(to_client_proc, to_client);
-	MemoryContextSwitchTo(oldcontext);
-
-	ClientEncoding = &pg_enc2name_tbl[encoding];
-	ToServerConvProc = to_server;
-	ToClientConvProc = to_client;
-
-	return 0;
 }
 
 /*
  * Initialize client encoding if necessary.
- *		called from InitPostgres() once during backend starting up.
+ *		called from InitPostgres() once during backend startup.
  */
 void
 InitializeClientEncoding(void)
@@ -195,7 +244,8 @@ InitializeClientEncoding(void)
 }
 
 /*
- * returns the current client encoding */
+ * returns the current client encoding
+ */
 int
 pg_get_client_encoding(void)
 {
@@ -220,8 +270,14 @@ pg_get_client_encoding_name(void)
  * it's taken from pg_catalog schema. If it even is not in the schema,
  * warn and return src.
  *
+ * If conversion occurs, a palloc'd null-terminated string is returned.
  * In the case of no conversion, src is returned.
  *
+ * CAUTION: although the presence of a length argument means that callers
+ * can pass non-null-terminated strings, care is required because the same
+ * string will be passed back if no conversion occurs.	Such callers *must*
+ * check whether result == src and handle that case differently.
+ *
  * Note: we try to avoid raising error, since that could get us into
  * infinite recursion when this function is invoked during error message
  * sending.  It should be OK to raise error for overlength strings though,
@@ -313,7 +369,7 @@ pg_convert_to(PG_FUNCTION_ARGS)
 	result = DirectFunctionCall3(pg_convert, string,
 								 src_encoding_name, dest_encoding_name);
 
-	PG_RETURN_BYTEA_P(result);
+	PG_RETURN_DATUM(result);
 }
 
 /*
@@ -340,7 +396,7 @@ pg_convert_from(PG_FUNCTION_ARGS)
 	 * in this case it will be because we've told pg_convert to return one
 	 * that is valid as text in the current database encoding.
 	 */
-	PG_RETURN_TEXT_P(result);
+	PG_RETURN_DATUM(result);
 }
 
 /*
@@ -380,8 +436,6 @@ pg_convert(PG_FUNCTION_ARGS)
 	*(str + len) = '\0';
 
 	result = pg_do_encoding_conversion(str, len, src_encoding, dest_encoding);
-	if (result == NULL)
-		elog(ERROR, "encoding conversion failed");
 
 	/*
 	 * build bytea data type structure.
@@ -506,10 +560,9 @@ pg_server_to_client(const char *s, int len)
 /*
  *	Perform default encoding conversion using cached FmgrInfo. Since
  *	this function does not access database at all, it is safe to call
- *	outside transactions. Explicit setting client encoding required
- *	before calling this function. Otherwise no conversion is
- *	performed.
-*/
+ *	outside transactions.  If the conversion has not been set up by
+ *	SetClientEncoding(), no conversion is performed.
+ */
 static char *
 perform_default_encoding_conversion(const char *src, int len, bool is_client_to_server)
 {
@@ -555,6 +608,127 @@ perform_default_encoding_conversion(const char *src, int len, bool is_client_to_
 	return result;
 }
 
+
+
+#ifdef USE_WIDE_UPPER_LOWER
+
+/*
+ * wchar2char --- convert wide characters to multibyte format
+ *
+ * This has the same API as the standard wcstombs() function; in particular,
+ * tolen is the maximum number of bytes to store at *to, and *from must be
+ * zero-terminated.  The output will be zero-terminated iff there is room.
+ */
+size_t
+wchar2char(char *to, const wchar_t *from, size_t tolen)
+{
+	size_t		result;
+
+	if (tolen == 0)
+		return 0;
+
+#ifdef WIN32
+
+	/*
+	 * On Windows, the "Unicode" locales assume UTF16 not UTF8 encoding, and
+	 * for some reason mbstowcs and wcstombs won't do this for us, so we use
+	 * MultiByteToWideChar().
+	 */
+	if (GetDatabaseEncoding() == PG_UTF8)
+	{
+		result = WideCharToMultiByte(CP_UTF8, 0, from, -1, to, tolen,
+									 NULL, NULL);
+		/* A zero return is failure */
+		if (result <= 0)
+			result = -1;
+		else
+		{
+			Assert(result <= tolen);
+			/* Microsoft counts the zero terminator in the result */
+			result--;
+		}
+	}
+	else
+#endif   /* WIN32 */
+	{
+		Assert(!lc_ctype_is_c());
+		result = wcstombs(to, from, tolen);
+	}
+	return result;
+}
+
+/*
+ * char2wchar --- convert multibyte characters to wide characters
+ *
+ * This has almost the API of mbstowcs(), except that *from need not be
+ * null-terminated; instead, the number of input bytes is specified as
+ * fromlen.  Also, we ereport() rather than returning -1 for invalid
+ * input encoding.	tolen is the maximum number of wchar_t's to store at *to.
+ * The output will be zero-terminated iff there is room.
+ */
+size_t
+char2wchar(wchar_t *to, size_t tolen, const char *from, size_t fromlen)
+{
+	size_t		result;
+
+	if (tolen == 0)
+		return 0;
+
+#ifdef WIN32
+	/* See WIN32 "Unicode" comment above */
+	if (GetDatabaseEncoding() == PG_UTF8)
+	{
+		/* Win32 API does not work for zero-length input */
+		if (fromlen == 0)
+			result = 0;
+		else
+		{
+			result = MultiByteToWideChar(CP_UTF8, 0, from, fromlen, to, tolen - 1);
+			/* A zero return is failure */
+			if (result == 0)
+				result = -1;
+		}
+
+		if (result != -1)
+		{
+			Assert(result < tolen);
+			/* Append trailing null wchar (MultiByteToWideChar() does not) */
+			to[result] = 0;
+		}
+	}
+	else
+#endif   /* WIN32 */
+	{
+		/* mbstowcs requires ending '\0' */
+		char	   *str = pnstrdup(from, fromlen);
+
+		Assert(!lc_ctype_is_c());
+		result = mbstowcs(to, str, tolen);
+		pfree(str);
+	}
+
+	if (result == -1)
+	{
+		/*
+		 * Invalid multibyte character encountered.  We try to give a useful
+		 * error message by letting pg_verifymbstr check the string.  But it's
+		 * possible that the string is OK to us, and not OK to mbstowcs ---
+		 * this suggests that the LC_CTYPE locale is different from the
+		 * database encoding.  Give a generic error message if verifymbstr
+		 * can't find anything wrong.
+		 */
+		pg_verifymbstr(from, fromlen, false);	/* might not return */
+		/* but if it does ... */
+		ereport(ERROR,
+				(errcode(ERRCODE_CHARACTER_NOT_IN_REPERTOIRE),
+				 errmsg("invalid multibyte character for locale"),
+				 errhint("The server's LC_CTYPE locale is probably incompatible with the database encoding.")));
+	}
+
+	return result;
+}
+#endif
+
 /* convert a multibyte string to a wchar */
 int
 pg_mb2wchar(const char *from, pg_wchar *to)
@@ -577,14 +751,14 @@ pg_encoding_mb2wchar_with_len(int encoding,
 	return (*pg_wchar_table[encoding].mb2wchar_with_len) ((const unsigned char *) from, to, len);
 }
 
-/* returns the byte length of a multibyte word */
+/* returns the byte length of a multibyte character */
 int
 pg_mblen(const char *mbstr)
 {
 	return ((*pg_wchar_table[DatabaseEncoding->encoding].mblen) ((const unsigned char *) mbstr));
 }
 
-/* returns the display length of a multibyte word */
+/* returns the display length of a multibyte character */
 int
 pg_dsplen(const char *mbstr)
 {
@@ -634,23 +808,37 @@ pg_mbstrlen_with_len(const char *mbstr, int limit)
 
 /*
  * returns the byte length of a multibyte string
- * (not necessarily  NULL terminated)
+ * (not necessarily NULL terminated)
  * that is no longer than limit.
- * this function does not break multibyte word boundary.
+ * this function does not break multibyte character boundary.
  */
 int
 pg_mbcliplen(const char *mbstr, int len, int limit)
 {
+	return pg_encoding_mbcliplen(DatabaseEncoding->encoding, mbstr,
+								 len, limit);
+}
+
+/*
+ * pg_mbcliplen with specified encoding
+ */
+int
+pg_encoding_mbcliplen(int encoding, const char *mbstr,
+					  int len, int limit)
+{
+	mblen_converter mblen_fn;
 	int			clen = 0;
 	int			l;
 
 	/* optimization for single byte encoding */
-	if (pg_database_encoding_max_length() == 1)
+	if (pg_encoding_max_length(encoding) == 1)
 		return cliplen(mbstr, len, limit);
 
+	mblen_fn = pg_wchar_table[encoding].mblen;
+
 	while (len > 0 && *mbstr)
 	{
-		l = pg_mblen(mbstr);
+		l = (*mblen_fn) ((const unsigned char *) mbstr);
 		if ((clen + l) > limit)
 			break;
 		clen += l;
@@ -664,7 +852,8 @@ pg_mbcliplen(const char *mbstr, int len, int limit)
 
 /*
  * Similar to pg_mbcliplen except the limit parameter specifies the
- * character length, not the byte length.  */
+ * character length, not the byte length.
+ */
 int
 pg_mbcharcliplen(const char *mbstr, int len, int limit)
 {
@@ -689,6 +878,18 @@ pg_mbcharcliplen(const char *mbstr, int len, int limit)
 	return clen;
 }
 
+/* mbcliplen for any single-byte encoding */
+static int
+cliplen(const char *str, int len, int limit)
+{
+	int			l = 0;
+
+	len = Min(len, limit);
+	while (l < len && str[l])
+		l++;
+	return l;
+}
+
 void
 SetDatabaseEncoding(int encoding)
 {
@@ -699,10 +900,44 @@ SetDatabaseEncoding(int encoding)
 	Assert(DatabaseEncoding->encoding == encoding);
 }
 
+/*
+ * Bind gettext to the codeset equivalent with the database encoding.
+ */
 void
-SetDefaultClientEncoding(void)
+pg_bind_textdomain_codeset(const char *domainname)
 {
-	ClientEncoding = &pg_enc2name_tbl[GetDatabaseEncoding()];
+#if defined(ENABLE_NLS)
+	int			encoding = GetDatabaseEncoding();
+	int			i;
+
+	/*
+	 * gettext() uses the codeset specified by LC_CTYPE by default, so if that
+	 * matches the database encoding we don't need to do anything. In CREATE
+	 * DATABASE, we enforce or trust that the locale's codeset matches
+	 * database encoding, except for the C locale. In C locale, we bind
+	 * gettext() explicitly to the right codeset.
+	 *
+	 * On Windows, though, gettext() tends to get confused so we always bind
+	 * it.
+	 */
+#ifndef WIN32
+	const char *ctype = setlocale(LC_CTYPE, NULL);
+
+	if (pg_strcasecmp(ctype, "C") != 0 && pg_strcasecmp(ctype, "POSIX") != 0)
+		return;
+#endif
+
+	for (i = 0; pg_enc2gettext_tbl[i].name != NULL; i++)
+	{
+		if (pg_enc2gettext_tbl[i].encoding == encoding)
+		{
+			if (bind_textdomain_codeset(domainname,
+										pg_enc2gettext_tbl[i].name) == NULL)
+				elog(LOG, "bind_textdomain_codeset failed");
+			break;
+		}
+	}
+#endif
 }
 
 int
@@ -732,17 +967,3 @@ pg_client_encoding(PG_FUNCTION_ARGS)
 	Assert(ClientEncoding);
 	return DirectFunctionCall1(namein, CStringGetDatum(ClientEncoding->name));
 }
-
-static int
-cliplen(const char *str, int len, int limit)
-{
-	int			l = 0;
-	const char *s;
-
-	for (s = str; *s; s++, l++)
-	{
-		if (l >= len || l >= limit)
-			return l;
-	}
-	return (s - str);
-}
