@@ -7,11 +7,11 @@
  * type.
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/mmgr/aset.c,v 1.53 2003/09/13 22:25:38 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/mmgr/aset.c,v 1.79 2009/06/11 14:49:06 momjian Exp $
  *
  * NOTE:
  *	This is a new (Feb. 05, 1999) implementation of the allocation set
@@ -85,7 +85,7 @@
  * CAUTION: ALLOC_MINBITS must be large enough so that
  * 1<<ALLOC_MINBITS is at least MAXALIGN,
  * or we may fail to align the smallest chunks adequately.
- * 16-byte alignment is enough on all currently known machines.
+ * 8-byte alignment is enough on all currently known machines.
  *
  * With the current parameters, request sizes up to 8K are treated as chunks,
  * larger requests go into dedicated blocks.  Change ALLOCSET_NUM_FREELISTS
@@ -93,8 +93,8 @@
  *--------------------
  */
 
-#define ALLOC_MINBITS		4	/* smallest chunk size is 16 bytes */
-#define ALLOCSET_NUM_FREELISTS	10
+#define ALLOC_MINBITS		3	/* smallest chunk size is 8 bytes */
+#define ALLOCSET_NUM_FREELISTS	11
 #define ALLOC_CHUNK_LIMIT	(1 << (ALLOCSET_NUM_FREELISTS-1+ALLOC_MINBITS))
 /* Size of largest chunk that we use a fixed size for */
 
@@ -123,6 +123,12 @@ typedef void *AllocPointer;
 
 /*
  * AllocSetContext is our standard implementation of MemoryContext.
+ *
+ * Note: isReset means there is nothing for AllocSetReset to do.  This is
+ * different from the aset being physically empty (empty blocks list) because
+ * we may still have a keeper block.  It's also different from the set being
+ * logically empty, because we don't attempt to detect pfree'ing the last
+ * active chunk.
  */
 typedef struct AllocSetContext
 {
@@ -130,11 +136,13 @@ typedef struct AllocSetContext
 	/* Info about storage allocated in this context: */
 	AllocBlock	blocks;			/* head of list of blocks in this set */
 	AllocChunk	freelist[ALLOCSET_NUM_FREELISTS];		/* free chunk lists */
+	bool		isReset;		/* T = no space alloced since last reset */
 	/* Allocation parameters for this context: */
 	Size		initBlockSize;	/* initial block size */
 	Size		maxBlockSize;	/* maximum block size */
-	AllocBlock	keeper;			/* if not NULL, keep this block over
-								 * resets */
+	Size		nextBlockSize;	/* next block size to allocate */
+	Size		allocChunkLimit;	/* effective chunk size limit */
+	AllocBlock	keeper;			/* if not NULL, keep this block over resets */
 } AllocSetContext;
 
 typedef AllocSetContext *AllocSet;
@@ -205,7 +213,8 @@ static void AllocSetInit(MemoryContext context);
 static void AllocSetReset(MemoryContext context);
 static void AllocSetDelete(MemoryContext context);
 static Size AllocSetGetChunkSpace(MemoryContext context, void *pointer);
-static void AllocSetStats(MemoryContext context);
+static bool AllocSetIsEmpty(MemoryContext context);
+static void AllocSetStats(MemoryContext context, int level);
 
 #ifdef MEMORY_CONTEXT_CHECKING
 static void AllocSetCheck(MemoryContext context);
@@ -222,6 +231,7 @@ static MemoryContextMethods AllocSetMethods = {
 	AllocSetReset,
 	AllocSetDelete,
 	AllocSetGetChunkSpace,
+	AllocSetIsEmpty,
 	AllocSetStats
 #ifdef MEMORY_CONTEXT_CHECKING
 	,AllocSetCheck
@@ -272,6 +282,31 @@ AllocSetFreeIndex(Size size)
 	return idx;
 }
 
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+
+/*
+ * Fill a just-allocated piece of memory with "random" data.  It's not really
+ * very random, just a repeating sequence with a length that's prime.  What
+ * we mainly want out of it is to have a good probability that two palloc's
+ * of the same number of bytes start out containing different data.
+ */
+static void
+randomize_mem(char *ptr, size_t size)
+{
+	static int	save_ctr = 1;
+	int			ctr;
+
+	ctr = save_ctr;
+	while (size-- > 0)
+	{
+		*ptr++ = ctr;
+		if (++ctr > 251)
+			ctr = 1;
+	}
+	save_ctr = ctr;
+}
+#endif   /* RANDOMIZE_ALLOCATED_MEMORY */
+
 
 /*
  * Public routines
@@ -317,6 +352,21 @@ AllocSetContextCreate(MemoryContext parent,
 		maxBlockSize = initBlockSize;
 	context->initBlockSize = initBlockSize;
 	context->maxBlockSize = maxBlockSize;
+	context->nextBlockSize = initBlockSize;
+
+	/*
+	 * Compute the allocation chunk size limit for this context.  It can't be
+	 * more than ALLOC_CHUNK_LIMIT because of the fixed number of freelists.
+	 * If maxBlockSize is small then requests exceeding the maxBlockSize
+	 * should be treated as large chunks, too.	We have to have
+	 * allocChunkLimit a power of two, because the requested and
+	 * actually-allocated sizes of any chunk must be on the same side of the
+	 * limit, else we get confused about whether the chunk is "big".
+	 */
+	context->allocChunkLimit = ALLOC_CHUNK_LIMIT;
+	while (context->allocChunkLimit >
+		   (Size) (maxBlockSize - ALLOC_BLOCKHDRSZ - ALLOC_CHUNKHDRSZ))
+		context->allocChunkLimit >>= 1;
 
 	/*
 	 * Grab always-allocated space, if requested
@@ -333,8 +383,8 @@ AllocSetContextCreate(MemoryContext parent,
 			ereport(ERROR,
 					(errcode(ERRCODE_OUT_OF_MEMORY),
 					 errmsg("out of memory"),
-				errdetail("Failed while creating memory context \"%s\".",
-						  name)));
+					 errdetail("Failed while creating memory context \"%s\".",
+							   name)));
 		}
 		block->aset = context;
 		block->freeptr = ((char *) block) + ALLOC_BLOCKHDRSZ;
@@ -345,6 +395,8 @@ AllocSetContextCreate(MemoryContext parent,
 		context->keeper = block;
 	}
 
+	context->isReset = true;
+
 	return (MemoryContext) context;
 }
 
@@ -384,17 +436,24 @@ static void
 AllocSetReset(MemoryContext context)
 {
 	AllocSet	set = (AllocSet) context;
-	AllocBlock	block = set->blocks;
+	AllocBlock	block;
 
 	AssertArg(AllocSetIsValid(set));
 
+	/* Nothing to do if no pallocs since startup or last reset */
+	if (set->isReset)
+		return;
+
 #ifdef MEMORY_CONTEXT_CHECKING
 	/* Check for corruption and leaks before freeing */
 	AllocSetCheck(context);
 #endif
 
 	/* Clear chunk freelists */
-	MemSet(set->freelist, 0, sizeof(set->freelist));
+	MemSetAligned(set->freelist, 0, sizeof(set->freelist));
+
+	block = set->blocks;
+
 	/* New blocks list is either empty or just the keeper block */
 	set->blocks = set->keeper;
 
@@ -425,6 +484,11 @@ AllocSetReset(MemoryContext context)
 		}
 		block = next;
 	}
+
+	/* Reset block size allocation sequence, too */
+	set->nextBlockSize = set->initBlockSize;
+
+	set->isReset = true;
 }
 
 /*
@@ -449,7 +513,7 @@ AllocSetDelete(MemoryContext context)
 #endif
 
 	/* Make it look empty, just in case... */
-	MemSet(set->freelist, 0, sizeof(set->freelist));
+	MemSetAligned(set->freelist, 0, sizeof(set->freelist));
 	set->blocks = NULL;
 	set->keeper = NULL;
 
@@ -477,7 +541,6 @@ AllocSetAlloc(MemoryContext context, Size size)
 	AllocSet	set = (AllocSet) context;
 	AllocBlock	block;
 	AllocChunk	chunk;
-	AllocChunk	priorfree;
 	int			fidx;
 	Size		chunk_size;
 	Size		blksize;
@@ -485,10 +548,10 @@ AllocSetAlloc(MemoryContext context, Size size)
 	AssertArg(AllocSetIsValid(set));
 
 	/*
-	 * If requested size exceeds maximum for chunks, allocate an entire
-	 * block for this request.
+	 * If requested size exceeds maximum for chunks, allocate an entire block
+	 * for this request.
 	 */
-	if (size > ALLOC_CHUNK_LIMIT)
+	if (size > set->allocChunkLimit)
 	{
 		chunk_size = MAXALIGN(size);
 		blksize = chunk_size + ALLOC_BLOCKHDRSZ + ALLOC_CHUNKHDRSZ;
@@ -514,10 +577,14 @@ AllocSetAlloc(MemoryContext context, Size size)
 		if (size < chunk_size)
 			((char *) AllocChunkGetPointer(chunk))[size] = 0x7E;
 #endif
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+		/* fill the allocated space with junk */
+		randomize_mem((char *) AllocChunkGetPointer(chunk), size);
+#endif
 
 		/*
-		 * Stick the new block underneath the active allocation block, so
-		 * that we don't lose the use of the space remaining therein.
+		 * Stick the new block underneath the active allocation block, so that
+		 * we don't lose the use of the space remaining therein.
 		 */
 		if (set->blocks != NULL)
 		{
@@ -530,34 +597,25 @@ AllocSetAlloc(MemoryContext context, Size size)
 			set->blocks = block;
 		}
 
+		set->isReset = false;
+
 		AllocAllocInfo(set, chunk);
 		return AllocChunkGetPointer(chunk);
 	}
 
 	/*
 	 * Request is small enough to be treated as a chunk.  Look in the
-	 * corresponding free list to see if there is a free chunk we could
-	 * reuse.
+	 * corresponding free list to see if there is a free chunk we could reuse.
+	 * If one is found, remove it from the free list, make it again a member
+	 * of the alloc set and return its data address.
 	 */
 	fidx = AllocSetFreeIndex(size);
-	priorfree = NULL;
-	for (chunk = set->freelist[fidx]; chunk; chunk = (AllocChunk) chunk->aset)
-	{
-		if (chunk->size >= size)
-			break;
-		priorfree = chunk;
-	}
-
-	/*
-	 * If one is found, remove it from the free list, make it again a
-	 * member of the alloc set and return its data address.
-	 */
+	chunk = set->freelist[fidx];
 	if (chunk != NULL)
 	{
-		if (priorfree == NULL)
-			set->freelist[fidx] = (AllocChunk) chunk->aset;
-		else
-			priorfree->aset = chunk->aset;
+		Assert(chunk->size >= size);
+
+		set->freelist[fidx] = (AllocChunk) chunk->aset;
 
 		chunk->aset = (void *) set;
 
@@ -567,6 +625,13 @@ AllocSetAlloc(MemoryContext context, Size size)
 		if (size < chunk->size)
 			((char *) AllocChunkGetPointer(chunk))[size] = 0x7E;
 #endif
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+		/* fill the allocated space with junk */
+		randomize_mem((char *) AllocChunkGetPointer(chunk), size);
+#endif
+
+		/* isReset must be false already */
+		Assert(!set->isReset);
 
 		AllocAllocInfo(set, chunk);
 		return AllocChunkGetPointer(chunk);
@@ -575,12 +640,12 @@ AllocSetAlloc(MemoryContext context, Size size)
 	/*
 	 * Choose the actual chunk size to allocate.
 	 */
-	chunk_size = 1 << (fidx + ALLOC_MINBITS);
+	chunk_size = (1 << ALLOC_MINBITS) << fidx;
 	Assert(chunk_size >= size);
 
 	/*
-	 * If there is enough room in the active allocation block, we will put
-	 * the chunk into that block.  Else must start a new one.
+	 * If there is enough room in the active allocation block, we will put the
+	 * chunk into that block.  Else must start a new one.
 	 */
 	if ((block = set->blocks) != NULL)
 	{
@@ -589,16 +654,16 @@ AllocSetAlloc(MemoryContext context, Size size)
 		if (availspace < (chunk_size + ALLOC_CHUNKHDRSZ))
 		{
 			/*
-			 * The existing active (top) block does not have enough room
-			 * for the requested allocation, but it might still have a
-			 * useful amount of space in it.  Once we push it down in the
-			 * block list, we'll never try to allocate more space from it.
-			 * So, before we do that, carve up its free space into chunks
-			 * that we can put on the set's freelists.
+			 * The existing active (top) block does not have enough room for
+			 * the requested allocation, but it might still have a useful
+			 * amount of space in it.  Once we push it down in the block list,
+			 * we'll never try to allocate more space from it. So, before we
+			 * do that, carve up its free space into chunks that we can put on
+			 * the set's freelists.
 			 *
 			 * Because we can only get here when there's less than
-			 * ALLOC_CHUNK_LIMIT left in the block, this loop cannot
-			 * iterate more than ALLOCSET_NUM_FREELISTS-1 times.
+			 * ALLOC_CHUNK_LIMIT left in the block, this loop cannot iterate
+			 * more than ALLOCSET_NUM_FREELISTS-1 times.
 			 */
 			while (availspace >= ((1 << ALLOC_MINBITS) + ALLOC_CHUNKHDRSZ))
 			{
@@ -606,10 +671,9 @@ AllocSetAlloc(MemoryContext context, Size size)
 				int			a_fidx = AllocSetFreeIndex(availchunk);
 
 				/*
-				 * In most cases, we'll get back the index of the next
-				 * larger freelist than the one we need to put this chunk
-				 * on.	The exception is when availchunk is exactly a
-				 * power of 2.
+				 * In most cases, we'll get back the index of the next larger
+				 * freelist than the one we need to put this chunk on.	The
+				 * exception is when availchunk is exactly a power of 2.
 				 */
 				if (availchunk != (1 << (a_fidx + ALLOC_MINBITS)))
 				{
@@ -643,32 +707,18 @@ AllocSetAlloc(MemoryContext context, Size size)
 	{
 		Size		required_size;
 
-		if (set->blocks == NULL)
-		{
-			/* First block of the alloc set, use initBlockSize */
-			blksize = set->initBlockSize;
-		}
-		else
-		{
-			/*
-			 * Use first power of 2 that is larger than previous block,
-			 * but not more than the allowed limit.  (We don't simply double
-			 * the prior block size, because in some cases this could be a
-			 * funny size, eg if very first allocation was for an odd-sized
-			 * large chunk.)
-			 */
-			Size	pblksize = set->blocks->endptr - ((char *) set->blocks);
-
-			blksize = set->initBlockSize;
-			while (blksize <= pblksize)
-				blksize <<= 1;
-			if (blksize > set->maxBlockSize)
-				blksize = set->maxBlockSize;
-		}
+		/*
+		 * The first such block has size initBlockSize, and we double the
+		 * space in each succeeding block, but not more than maxBlockSize.
+		 */
+		blksize = set->nextBlockSize;
+		set->nextBlockSize <<= 1;
+		if (set->nextBlockSize > set->maxBlockSize)
+			set->nextBlockSize = set->maxBlockSize;
 
 		/*
-		 * If initBlockSize is less than ALLOC_CHUNK_LIMIT, we could need
-		 * more space... but try to keep it a power of 2.
+		 * If initBlockSize is less than ALLOC_CHUNK_LIMIT, we could need more
+		 * space... but try to keep it a power of 2.
 		 */
 		required_size = chunk_size + ALLOC_BLOCKHDRSZ + ALLOC_CHUNKHDRSZ;
 		while (blksize < required_size)
@@ -678,9 +728,8 @@ AllocSetAlloc(MemoryContext context, Size size)
 		block = (AllocBlock) malloc(blksize);
 
 		/*
-		 * We could be asking for pretty big blocks here, so cope if
-		 * malloc fails.  But give up if there's less than a meg or so
-		 * available...
+		 * We could be asking for pretty big blocks here, so cope if malloc
+		 * fails.  But give up if there's less than a meg or so available...
 		 */
 		while (block == NULL && blksize > 1024 * 1024)
 		{
@@ -705,19 +754,16 @@ AllocSetAlloc(MemoryContext context, Size size)
 		block->endptr = ((char *) block) + blksize;
 
 		/*
-		 * If this is the first block of the set, make it the "keeper"
-		 * block. Formerly, a keeper block could only be created during
-		 * context creation, but allowing it to happen here lets us have
-		 * fast reset cycling even for contexts created with
-		 * minContextSize = 0; that way we don't have to force space to be
-		 * allocated in contexts that might never need any space.  Don't
-		 * mark an oversize block as a keeper, however.
+		 * If this is the first block of the set, make it the "keeper" block.
+		 * Formerly, a keeper block could only be created during context
+		 * creation, but allowing it to happen here lets us have fast reset
+		 * cycling even for contexts created with minContextSize = 0; that way
+		 * we don't have to force space to be allocated in contexts that might
+		 * never need any space.  Don't mark an oversize block as a keeper,
+		 * however.
 		 */
-		if (set->blocks == NULL && blksize == set->initBlockSize)
-		{
-			Assert(set->keeper == NULL);
+		if (set->keeper == NULL && blksize == set->initBlockSize)
 			set->keeper = block;
-		}
 
 		block->next = set->blocks;
 		set->blocks = block;
@@ -739,6 +785,12 @@ AllocSetAlloc(MemoryContext context, Size size)
 	if (size < chunk->size)
 		((char *) AllocChunkGetPointer(chunk))[size] = 0x7E;
 #endif
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+	/* fill the allocated space with junk */
+	randomize_mem((char *) AllocChunkGetPointer(chunk), size);
+#endif
+
+	set->isReset = false;
 
 	AllocAllocInfo(set, chunk);
 	return AllocChunkGetPointer(chunk);
@@ -764,7 +816,7 @@ AllocSetFree(MemoryContext context, void *pointer)
 				 set->header.name, chunk);
 #endif
 
-	if (chunk->size > ALLOC_CHUNK_LIMIT)
+	if (chunk->size > set->allocChunkLimit)
 	{
 		/*
 		 * Big chunks are certain to have been allocated as single-chunk
@@ -838,14 +890,24 @@ AllocSetRealloc(MemoryContext context, void *pointer, Size size)
 				 set->header.name, chunk);
 #endif
 
+	/* isReset must be false already */
+	Assert(!set->isReset);
+
 	/*
 	 * Chunk sizes are aligned to power of 2 in AllocSetAlloc(). Maybe the
-	 * allocated area already is >= the new size.  (In particular, we
-	 * always fall out here if the requested size is a decrease.)
+	 * allocated area already is >= the new size.  (In particular, we always
+	 * fall out here if the requested size is a decrease.)
 	 */
 	if (oldsize >= size)
 	{
 #ifdef MEMORY_CONTEXT_CHECKING
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+		/* We can only fill the extra space if we know the prior request */
+		if (size > chunk->requested_size)
+			randomize_mem((char *) AllocChunkGetPointer(chunk) + chunk->requested_size,
+						  size - chunk->requested_size);
+#endif
+
 		chunk->requested_size = size;
 		/* set mark to catch clobber of "unused" space */
 		if (size < oldsize)
@@ -854,10 +916,10 @@ AllocSetRealloc(MemoryContext context, void *pointer, Size size)
 		return pointer;
 	}
 
-	if (oldsize > ALLOC_CHUNK_LIMIT)
+	if (oldsize > set->allocChunkLimit)
 	{
 		/*
-		 * The chunk must been allocated as a single-chunk block.  Find
+		 * The chunk must have been allocated as a single-chunk block.	Find
 		 * the containing block and use realloc() to make it bigger with
 		 * minimum space wastage.
 		 */
@@ -903,6 +965,12 @@ AllocSetRealloc(MemoryContext context, void *pointer, Size size)
 		chunk->size = chksize;
 
 #ifdef MEMORY_CONTEXT_CHECKING
+#ifdef RANDOMIZE_ALLOCATED_MEMORY
+		/* We can only fill the extra space if we know the prior request */
+		randomize_mem((char *) AllocChunkGetPointer(chunk) + chunk->requested_size,
+					  size - chunk->requested_size);
+#endif
+
 		chunk->requested_size = size;
 		/* set mark to catch clobber of "unused" space */
 		if (size < chunk->size)
@@ -914,57 +982,18 @@ AllocSetRealloc(MemoryContext context, void *pointer, Size size)
 	else
 	{
 		/*
-		 * Small-chunk case.  If the chunk is the last one in its block,
-		 * there might be enough free space after it that we can just
-		 * enlarge the chunk in-place.	It's relatively painful to find
-		 * the containing block in the general case, but we can detect
-		 * last-ness quite cheaply for the typical case where the chunk is
-		 * in the active (topmost) allocation block.  (At least with the
-		 * regression tests and code as of 1/2001, realloc'ing the last
-		 * chunk of a non-topmost block hardly ever happens, so it's not
-		 * worth scanning the block list to catch that case.)
-		 *
-		 * NOTE: must be careful not to create a chunk of a size that
-		 * AllocSetAlloc would not create, else we'll get confused later.
+		 * Small-chunk case.  We just do this by brute force, ie, allocate a
+		 * new chunk and copy the data.  Since we know the existing data isn't
+		 * huge, this won't involve any great memcpy expense, so it's not
+		 * worth being smarter.  (At one time we tried to avoid memcpy when it
+		 * was possible to enlarge the chunk in-place, but that turns out to
+		 * misbehave unpleasantly for repeated cycles of
+		 * palloc/repalloc/pfree: the eventually freed chunks go into the
+		 * wrong freelist for the next initial palloc request, and so we leak
+		 * memory indefinitely.  See pgsql-hackers archives for 2007-08-11.)
 		 */
 		AllocPointer newPointer;
 
-		if (size <= ALLOC_CHUNK_LIMIT)
-		{
-			AllocBlock	block = set->blocks;
-			char	   *chunk_end;
-
-			chunk_end = (char *) chunk + (oldsize + ALLOC_CHUNKHDRSZ);
-			if (chunk_end == block->freeptr)
-			{
-				/* OK, it's last in block ... is there room? */
-				Size		freespace = block->endptr - block->freeptr;
-				int			fidx;
-				Size		newsize;
-				Size		delta;
-
-				fidx = AllocSetFreeIndex(size);
-				newsize = 1 << (fidx + ALLOC_MINBITS);
-				Assert(newsize >= oldsize);
-				delta = newsize - oldsize;
-				if (freespace >= delta)
-				{
-					/* Yes, so just enlarge the chunk. */
-					block->freeptr += delta;
-					chunk->size += delta;
-#ifdef MEMORY_CONTEXT_CHECKING
-					chunk->requested_size = size;
-					/* set mark to catch clobber of "unused" space */
-					if (size < chunk->size)
-						((char *) pointer)[size] = 0x7E;
-#endif
-					return pointer;
-				}
-			}
-		}
-
-		/* Normal small-chunk case: just do it by brute force. */
-
 		/* allocate new chunk */
 		newPointer = AllocSetAlloc((MemoryContext) set, size);
 
@@ -991,12 +1020,32 @@ AllocSetGetChunkSpace(MemoryContext context, void *pointer)
 	return chunk->size + ALLOC_CHUNKHDRSZ;
 }
 
+/*
+ * AllocSetIsEmpty
+ *		Is an allocset empty of any allocated space?
+ */
+static bool
+AllocSetIsEmpty(MemoryContext context)
+{
+	AllocSet	set = (AllocSet) context;
+
+	/*
+	 * For now, we say "empty" only if the context is new or just reset. We
+	 * could examine the freelists to determine if all space has been freed,
+	 * but it's not really worth the trouble for present uses of this
+	 * functionality.
+	 */
+	if (set->isReset)
+		return true;
+	return false;
+}
+
 /*
  * AllocSetStats
  *		Displays stats about memory consumption of an allocset.
  */
 static void
-AllocSetStats(MemoryContext context)
+AllocSetStats(MemoryContext context, int level)
 {
 	AllocSet	set = (AllocSet) context;
 	long		nblocks = 0;
@@ -1006,6 +1055,7 @@ AllocSetStats(MemoryContext context)
 	AllocBlock	block;
 	AllocChunk	chunk;
 	int			fidx;
+	int			i;
 
 	for (block = set->blocks; block != NULL; block = block->next)
 	{
@@ -1022,8 +1072,12 @@ AllocSetStats(MemoryContext context)
 			freespace += chunk->size + ALLOC_CHUNKHDRSZ;
 		}
 	}
+
+	for (i = 0; i < level; i++)
+		fprintf(stderr, "  ");
+
 	fprintf(stderr,
-		"%s: %ld total in %ld blocks; %ld free (%ld chunks); %ld used\n",
+			"%s: %lu total in %ld blocks; %lu free (%ld chunks); %lu used\n",
 			set->header.name, totalspace, nblocks, freespace, nchunks,
 			totalspace - freespace);
 }
@@ -1088,15 +1142,15 @@ AllocSetCheck(MemoryContext context)
 					 name, (unsigned long) chsize, chunk, block);
 
 			/* single-chunk block? */
-			if (chsize > ALLOC_CHUNK_LIMIT &&
+			if (chsize > set->allocChunkLimit &&
 				chsize + ALLOC_CHUNKHDRSZ != blk_used)
 				elog(WARNING, "problem in alloc set %s: bad single-chunk %p in block %p",
 					 name, chunk, block);
 
 			/*
-			 * If chunk is allocated, check for correct aset pointer. (If
-			 * it's free, the aset is the freelist pointer, which we can't
-			 * check as easily...)
+			 * If chunk is allocated, check for correct aset pointer. (If it's
+			 * free, the aset is the freelist pointer, which we can't check as
+			 * easily...)
 			 */
 			if (dsize > 0 && chunk->aset != (void *) set)
 				elog(WARNING, "problem in alloc set %s: bogus aset link in block %p, chunk %p",
