@@ -1,6 +1,5 @@
 /* Processed by ecpg (regression mode) */
 /* These include files are added by the preprocessor */
-#include <ecpgtype.h>
 #include <ecpglib.h>
 #include <ecpgerrno.h>
 #include <sqlca.h>
@@ -75,19 +74,19 @@ main(void)
 		 
 	
 #line 52 "dt_test2.pgc"
- date  date1    ;
+ date date1 ;
  
 #line 53 "dt_test2.pgc"
- timestamp  ts1    ,  ts2    ;
+ timestamp ts1 , ts2 ;
  
 #line 54 "dt_test2.pgc"
- char * text    ;
+ char * text ;
  
 #line 55 "dt_test2.pgc"
- interval * i1    ;
+ interval * i1 ;
  
 #line 56 "dt_test2.pgc"
- date * dc    ;
+ date * dc ;
 /* exec sql end declare section */
 #line 57 "dt_test2.pgc"
 
@@ -140,6 +139,7 @@ main(void)
 					printf("TS[%d,%d]: %s\n",
 						i, j, errno ? "-" : text);
 				free(text);
+				free(t);
 			}
 		}
 	}
@@ -170,6 +170,7 @@ main(void)
 		printf("interval_copy[%d]: %s\n", i, text ? text : "-");
 		free(text);
 		PGTYPESinterval_free(ic);
+		PGTYPESinterval_free(i1);
 	}
 
 	return (0);
